MODULE tree

! Here we define the data type NODE which will contain
  TYPE node
    INTEGER               :: level         ! level in tree: top=m, bottom=1
    INTEGER               :: lags(26)      ! lags for node are stored here
  DOUBLE PRECISION        :: phi(26)       ! coefficients for node
  DOUBLE PRECISION        :: v             ! MSE (white noise) for node
  DOUBLE PRECISION        :: eps(1:10100)  ! the epsilons for the node
  DOUBLE PRECISION        :: eta(-99:10000)! the etas for the node
    TYPE (node), POINTER  :: reg, star     ! pointers to the regular and
  END TYPE node                            ! starred subnodes one level down

! These will contain the end results
  DOUBLE PRECISION        :: topphi(26), topv, acvf(1000)

! Other globals
  INTEGER                       :: n, m, method
  DOUBLE PRECISION, ALLOCATABLE :: x(:)
  CHARACTER                     :: stamp*4


CONTAINS

SUBROUTINE make_tree(original_lags)
  INTEGER, ALLOCATABLE       :: toplags(:)
  INTEGER, INTENT (IN)       :: original_lags(m)
  DOUBLE PRECISION           :: x(n)
  TYPE (node), POINTER       :: top_node

  NULLIFY (top_node)                  !associates top_node so we can use it
  ALLOCATE(toplags(m))
  toplags=original_lags

! now build the tree of node lags
  CALL build_node_tree(m,toplags,top_node)  

! now fill the tree, ie. get coeffts and MSE's of each node
  CALL fill_tree(top_node)

! node tree built, so print it
  CALL print_node_tree(top_node) 

! Likelihood calculation
  CALL likelihood(original_lags, topphi(1:m), topv)

  DEALLOCATE(toplags)

  RETURN
END SUBROUTINE make_tree

!***********************************************************************

RECURSIVE SUBROUTINE build_node_tree(lev,this_lags,this_node)
  INTEGER                 :: i, lev, this_lags(26)
  TYPE (node), POINTER    :: this_node  

! This routine will create a tree of nodes needed to subset Burg model. The
! level and lags are assigned to each node. Also the MSE of each node is
! initialized to be zero so that later we'll be able to check which nodes
! have not yet been filled.

! first time thru' with a fresh node; point to it & make its J and J*
! pointers point nowhere
  IF (.NOT. ASSOCIATED(this_node)) THEN
    ALLOCATE (this_node)
    this_node%level=lev  
    this_node%lags(1:lev)=this_lags(1:lev)
!   the check for an unfilled node will be that its MSE=0
    this_node%v=0           
    NULLIFY (this_node%reg) 
    NULLIFY (this_node%star)  
  END IF

! recursive call to routine with J lags; only if level>1
!  lev=this_node%level
  IF (this_node%level>1) THEN 
    this_lags(1:lev-1)=this_node%lags(1:lev-1)                   
    CALL build_node_tree(lev-1,this_lags,this_node%reg)
  END IF

! recursive call to routine with J* lags; only if level>1
!  lev=this_node%level
  IF (this_node%level>1) THEN  
    this_lags(1:lev-1)=(/(this_node%lags(lev)              &
                                 -this_node%lags(lev-i), i=1,lev-1)/)   
    CALL build_node_tree(lev-1,this_lags,this_node%star)  
  END IF

  RETURN
END SUBROUTINE build_node_tree
 
!***********************************************************************

RECURSIVE SUBROUTINE fill_tree(this_node)
  INTEGER               :: i
  TYPE (node), POINTER  :: this_node  

! Here we fill in the coeffts and MSE for each node

  IF (this_node%level>1) THEN
    CALL fill_tree(this_node%reg)
!   if reg node has not been filled, then fill it!
    IF (this_node%reg%v==0) THEN 
      CALL fill_node(this_node%reg)
    END IF

    CALL fill_tree(this_node%star)

!   if star node has not been filled, then fill it!
    IF (this_node%star%v==0) THEN 
      CALL fill_node(this_node%star) 
    END IF
  END IF

! now that whole tree is filled, we can fill top node
  IF (this_node%level==m) THEN
    CALL fill_node(this_node)
  END IF

  RETURN
END SUBROUTINE fill_tree

!***********************************************************************

SUBROUTINE fill_node(this_node)
  INTEGER                       :: i, km, lev, sum_range(2)
  DOUBLE PRECISION, ALLOCATABLE :: eps_J(:),eps_Js(:),eta_J(:),eta_Js(:)
  DOUBLE PRECISION              :: v_J, v_Js, phi_K, phi_Ks, top, sum_en
  DOUBLE PRECISION              :: phi_J(26), phi_Js(26), sum_e, sum_n 
  TYPE (node), POINTER          :: this_node  

! This is the routine where the real work of building the coeffts, MSEs,
! epsilons and etas is done

! initialize the precursors before applying algo to this node
  lev=this_node%level
  km=this_node%lags(lev)
  ALLOCATE (eps_J(1:n+km),eps_Js(1:n+km),eta_J(1-km:n),eta_Js(1-km:n))
  IF (lev==1) THEN 
    eps_J =0.0
    eps_Js=0.0
    eta_J =0.0
    eta_Js=0.0
    eps_J (1:n)=x(1:n)
    eps_Js(1:n)=x(1:n)
    eta_J (1:n)=x(1:n)
    eta_Js(1:n)=x(1:n)
    v_J   =SUM(x**2)/n
    v_Js  =v_J
  ELSE  !we're at a higher level, so use reg and star node info 
    phi_J (1:lev-1)=this_node%reg%phi (1:lev-1)
    phi_Js(1:lev-1)=this_node%star%phi(1:lev-1)
    eps_J  (1:n+km)=this_node%reg%eps (1:n+km)
    eps_Js (1:n+km)=this_node%star%eps(1:n+km)    
    eta_J  (1-km:n)=this_node%reg%eta (1-km:n)
    eta_Js (1-km:n)=this_node%star%eta(1-km:n)
    v_J   =this_node%reg%v
    v_Js  =this_node%star%v
  END IF

! Initial conditions set, now apply algo to this node
! First the reflection coefficients
  IF (method==1) THEN ! YuWa
    sum_range=(/1,n+km/)
    top=0
    DO i=1,n+km
      top=top+(eps_J(i)*eta_Js(i-km))
    END DO
    phi_K =top/(n*v_Js)
    phi_Ks=top/(n*v_J)
  ELSE ! Burg type
    sum_range=(/1+km,n/)
!   First get sum of squares and cross squares for eps and eta:
    sum_e=0; sum_n=0; sum_en=0
    DO i=km+1,n
      sum_en=sum_en + eps_J(i)*eta_Js(i-km)
      sum_e=sum_e + eps_J(i)**2
      sum_n=sum_n + eta_Js(i-km)**2  
    END DO
!   Now calculate reflection coefficients depending on the method:
    SELECT CASE (method)
      CASE (2)  ! Burg
        phi_K =v_J*(v_J+v_Js)*sum_en/(sum_n*v_J**2 + sum_e*v_Js**2)
        phi_Ks=v_Js*(v_J+v_Js)*sum_en/(sum_n*v_J**2 + sum_e*v_Js**2)
      CASE (3)  ! Morf
        phi_K =SQRT(v_J/(v_Js*sum_e*sum_n))*sum_en
        phi_Ks=SQRT(v_Js/(v_J*sum_e*sum_n))*sum_en
      CASE (4)  ! Nutt
        phi_K =2.0*v_J*sum_en/(v_Js*sum_e+v_J*sum_n)
        phi_Ks=v_Js*phi_K/v_J
    END SELECT    
  END IF

! Continue with remaining recursions - identical for all algo's
  this_node%phi(lev)=phi_K
  IF (lev > 1) THEN
    DO i=1,lev-1
      this_node%phi(i)=phi_J(i)-phi_K*phi_Js(lev-i)
    END DO
  END IF
  this_node%v=(1-phi_Ks*phi_K)*v_J

! The eta's & epsilon's for posterity:
  DO i=sum_range(1),sum_range(2)
    this_node%eps(i)   =eps_J(i)-phi_K*eta_Js(i-km)
    this_node%eta(i-km)=eta_J(i-km)-phi_K*eps_Js(i)
  END DO

  DEALLOCATE (eps_J,eps_Js,eta_J,eta_Js)

  RETURN
END SUBROUTINE fill_node

!***********************************************************************

RECURSIVE SUBROUTINE print_node_tree(this_node)
  INTEGER               :: i
  TYPE (node), POINTER  :: this_node  

! Here we print the info in the top node - can also make it print all nodes
! by removing the inmost IF THEN loop

  IF (ASSOCIATED(this_node)) THEN
    CALL print_node_tree(this_node%reg)
    IF (this_node%level==m) THEN
      PRINT*,"***************************************************"
      PRINT*,"The estimated subset ",stamp," AR coefficients are:"
      DO i=1,this_node%level
        PRINT 10, this_node%lags(i), this_node%phi(i)
 10     FORMAT(" Phi(",I2,"): ",F9.5)
      END DO
      PRINT*,"***************************************************"
      PRINT*, stamp," WN variance estimate : ",this_node%v
!     Store results for likelihood calcs
      topphi(1:m)=this_node%phi(1:m)
      topv=this_node%v
    END IF
    CALL print_node_tree(this_node%star)
    IF (this_node%level<m) DEALLOCATE (this_node)
  END IF
  RETURN
END SUBROUTINE print_node_tree

!***********************************************************************

SUBROUTINE likelihood(lag, phi, s2) 
! First computes ACVF of a subset AR model with m coeffts (phi) and lags 
! (lag), and sigma^2=s2, into acvf, lags 0 to n-1 acvf(0:n-1). Then it gets 
! -2 log  Likelihood for the vector of obs x:
! -2log L = n*log(2*pi*s2) + sum(log(r(j))) + resid_ss/s2
! using the innnovations algorithm 
  INTEGER    :: i, j, k, km, t
  INTEGER    :: lag(m), lags(m+1)
  DOUBLE PRECISION :: phi(m), phis(m+1), acvf(0:2*lag(m)), Ka(n,n)
  DOUBLE PRECISION :: A(lag(m)+1, lag(m)+1), b(lag(m)+1)
  DOUBLE PRECISION :: s, s2, pi, resid_ss, xh(n), loglike_wn, loglike_ss
  DOUBLE PRECISION :: cond_like, th(1:n-1,0:n-1), r(0:n-1), aicc

  km=lag(m)
  pi=3.141592654
  lags(1)    =0
  lags(2:m+1)=lag
  phis(1)    =-1.0
  phis(2:m+1)=phi  
  b=0.0
  b(1)=s2
! Solve system A*acvf(0:km) = b, to get acvf(0:km)
  DO i=1,km+1
    A(i,:)=0.0
  END DO
  DO k=0,km
    DO j=0,m
      A(k+1, ABS(k-lags(j+1))+1) = A(k+1, ABS(k-lags(j+1))+1) - phis(j+1)
    END DO
  END DO      
  CALL DLSARG(1+km, A, 1+km, b, 1, acvf(0:km))

! Now get acvf(km+1:2*km) via recursions
  DO k=km+1, 2*km
    acvf(k)=0.0
    DO j=1,m
      acvf(k)=acvf(k)+phi(j)*acvf(k-lag(j))
    END DO
  END DO

! Form K(.,.) as in (5.3.5)
  DO i=1,n
    Ka(i,:)=0.0
  END DO
  DO i=1,n
    DO j=1,n
      IF (i<=km .AND. j<=km) Ka(i,j)=acvf(abs(i-j))/s2
      IF (min(i,j)<=km .AND. km<max(i,j) .AND. max(i,j)<=2*km) THEN
        s=0
        DO k=1,m
          s=s+phi(k)*acvf(abs(lag(k)-abs(i-j)))
        END DO
        Ka(i,j)=(acvf(abs(i-j))-s)/s2
      END IF
      IF (min(i,j)>km .AND. i==j) Ka(i,j)=1.0
     END DO
  END DO
  
! Form (5.2.16) recursions
  th(:,0)=1.0
  r(0)=Ka(1,1)
  DO i=1,n-1
    DO k=0,i-1
      s=0
      DO j=0,k-1
        s=s+th(k,k-j)*th(i,i-j)*r(j)
      END DO
      th(i,i-k)=(Ka(i+1,k+1)-s)/r(k)
    END DO
    s=0
    DO j=0,i-1
      s=s+r(j)*th(i,i-j)**2
    END DO
    r(i)=Ka(i+1,i+1)-s
  END DO

! Build 1-step predictors (xh's), and get the resid_ss 
  xh(1)=0.0
  DO k=1,n-1
    IF (k < km) THEN 
      xh(k+1)=0.0
      DO j=1,k
        xh(k+1)=xh(k+1)+th(k,j)*(x(k+1-j)-xh(k+1-j))
      END DO
    ELSE ! k >= km
      xh(k+1)=0.0
      DO j=1,m
        xh(k+1)=xh(k+1)+phi(j)*x(k+1-lag(j))
      END DO
    END IF
  END DO
  s=0
  resid_ss=0
  DO j=1,n
    IF (r(j-1)<=0) THEN
      PRINT*, "### NON-CAUSAL MODEL ###"
      RETURN
    END IF
    s=s+log(r(j-1))
    resid_ss=resid_ss+(x(j)-xh(j))**2/r(j-1)
  END DO

! calculate cond. likelihood
! -2 log CL = (n-km)*log(s2)+(1/s2)sum_{km+1}^n (x_t-phi_k1*x_{t-k1}-...
!            -phi_km*x_{t-km})^2
  cond_like=0
  DO t=km+1,n  
    cond_like=cond_like+DOT_PRODUCT(phi,(/(x(t-lag(j)), j=1,m)/))
  END DO
  cond_like=(n-m)*log(s2)+cond_like/s2
 
! Finally: -2log Likelihood=loglike: wn means use WN variance estimate:
! SS means use RSS/n variance estimate 
  loglike_wn = n*log(2.0*pi*s2) + s + resid_ss/s2
  loglike_ss = n*log(2.0*pi*resid_ss/n) + s + FLOAT(n)
  PRINT *,"RSS/n WN variance estimate: ", resid_ss/n
  PRINT*,"-2 Log Like (",stamp,") :",loglike_wn
  PRINT*,"-2 Log Like (RSS/n):",loglike_ss
  aicc=loglike_ss+2.0*n*(m+1)/FLOAT(n-m-2)
  PRINT*,"AICC (RSS/n): ", aicc
!      PRINT*, " s ",s
!  PRINT *,"-2 Log Cond Like (YW WN): ", cond_like 
  PRINT*,"%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%"

  RETURN
END SUBROUTINE likelihood

END MODULE tree

!***********************************************************************
!***********************************************************************

PROGRAM bdt

! Author: A. Trindade, www.stat.ufl.edu/~trindade/
! This program subset models AR(p)'s using Burg type recursions
! with subset size <=26, km<100, and a max of 10,000 observations.
! For details refer to the paper: "Implementing Modified Burg Algorithms 
! in Multivariate Subset Autoregression", by the author. 

  USE tree
  INTEGER                                :: i, mc
  INTEGER, ALLOCATABLE, DIMENSION (:)    :: toplags
  DOUBLE PRECISION                       :: y(10000), mean
  CHARACTER                              :: h*24

  PRINT*, "%%%%%%%%%%%%%%%% Univariate SAR Modeling Program %%%%%%%%%%%%%%%"
! read in the time series
  20   write(*,*)
       write(*,22)
  22    format(5x,'File name of time series for modelling: ',$)
  23    h='                        '
        read(*,*) h
        IF (h=='a                       ') h='lynx10.tsm'
        open(3,file=h,status='old',err=20)
        i=1
  25    read(3,*,end=30) y(i)
        i=i+1
          if (i.eq.10002) then
           write(*,6665)
6665       format(3x,'DATA TRUNCATED AFTER FIRST 10000 OBSERVATIONS.')
           i=i-1
           goto 30
          endif
        goto 25
  30    n=i-1
        close(3)

! now mean-correct the obs
  mean=SUM(y)/n
  ALLOCATE (x(n))  
  x=(/(y(i),i=1,n)/)
  PRINT*, "Do you wish to mean-correct the observations (1=yes, 0=no)?"
  READ*, mc
  IF (mc==1) x=(/(y(i)-mean,i=1,n)/)
  PRINT*, "There are ",n," observations."
  PRINT*,"First obs is ",x(1)," last is ",x(n)

! Enter how many lags will be modeled
  PRINT*, "Enter the number of lags to be modeled (<27):"
  READ*, m
  ALLOCATE (toplags(m))

! read in the lags
  PRINT*, "Enter the lags:"
  READ*, (toplags(i), i=1,m)
! PRINT*,"The lags are: ", (toplags(i), i=1,m)
  
! Enter the method
  PRINT*, "Enter the method for obtaining the reflection coefficients:"
  PRINT*, "Yule-Walker (1), Burg (2), Vieira-Morf (3), Nuttall-Strand (4):"
  READ*, method
  SELECT CASE (method)
    CASE (1); stamp="YuWa"
    CASE (2); stamp="Burg"
    CASE (3); stamp="Morf"
    CASE (4); stamp="Nutt"
    CASE DEFAULT
      PRINT*,"Method not in range: ", method   
      STOP
  END SELECT 

  CALL make_tree(toplags)

END PROGRAM bdt      


